# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import colorama

from twine import __main__ as dunder_main


def test_exception_handling(monkeypatch):
    monkeypatch.setattr(sys, "argv", ["twine", "upload", "missing.whl"])
    message = "InvalidDistribution: Cannot find file (or expand pattern): 'missing.whl'"
    assert dunder_main.main() == colorama.Fore.RED + message + colorama.Style.RESET_ALL


def test_no_color_exception(monkeypatch):
    monkeypatch.setattr(sys, "argv", ["twine", "--no-color", "upload", "missing.whl"])
    message = "InvalidDistribution: Cannot find file (or expand pattern): 'missing.whl'"
    assert dunder_main.main() == message
